#Copyright 2009 Diego Duclos
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
import attribute
import math
import dataFolder
import copy
import itertools
import character
import item
import group
import re
import xml.dom
import sys
import traceback

slotTypes = ["high", "medium", "low", "rig", "subsystem"]

slotShortNameMap = {"high" : "hi",
                    "low" : "low",
                    "medium" : "med",
                    "rig" : "rig",
                    "subsystem" : "subsystem"}

slotNameMap = {'high' : 'hiSlots',
               'medium' : 'medSlots',
               'low' : 'lowSlots',
               'rig' : 'rigSlots',
               'subsystem' : 'maxSubSystems'}

modStates = {-1 : 'offline',
             0 : 'inactive',
             1: 'active',
             2: 'overheated'}

STATE_OFFLINE = -1
STATE_INACTIVE = 0
STATE_ACTIVE = 1
STATE_OVERLOADED = 2

defaultDamagePattern = {'thermal' : 0.25,
                        'kinetic' : 0.25,
                        'explosive' : 0.25,
                        'em' : 0.25}

PEAK_RECHARGE = -(math.sqrt(2) - 2 ) / 2

class fitting(object):
    """
    The class used to represent a fit.
    """
    def __init__(self, ship = None, modules = None, character = None, drones = None, extraGangModules = None, activeGangModules = None, projEffects = None, implants = None, boosters = None, factorInReload = False):
        '''
        Constructor
        @param ship: the ship this fit is using
        @param modules: a list of tuples, each containing a module and it's ammo
        @param character: the character using this fit
        '''
        self.ship = ship
        self.modules = modules or []
        self.drones = drones or {}
        self.character = character
        self.boosters = boosters or {}
        self.implants = implants or {}
        self.blockedModules = set()
        #Extra gang bonussess
        if extraGangModules != None:
            self.extraGangModules = extraGangModules
        else:
            self.extraGangModules = {}
            #Loop through the leadership skill group and add em if they have any effects related to leadership
            grp = group.getGroup(name = "Leadership")
            items = item.getItemsByGroup(group = grp)
            for i in items:
                if i.published == 1:
                    for effect in i.effects:
                        if "gang" in effect.type:
                            self.extraGangModules[i] = {"level" : 0}
                            break
                    
        #The gang bonus table is AUTOGENERATED, don't bother trying to set it
        self.gangModules = {}
        self.activeGangModules = activeGangModules or {}
        self.projEffects = projEffects or {}
        
        if ship:
            self.setShip(ship)
        
        self.factorInReload = factorInReload
        
    def setShip(self, ship):
        self.ship = ship
        ship.attributes["_maxLockedTargetsSkill"] = attribute.basicAttribute(self.ship, "_maxLockedTargetsSkill", 2)
                
        ship.attributes["_maxActiveDrones"] = attribute.basicAttribute(self.ship, "_maxActiveDrones", 0)
                
    def getNumModulesAttributeValue(self, attribute, value, minState = STATE_OFFLINE):
        '''
        Get the number of modules fitted to the ship that have a certain attribute set to a certain value
        @param attribute: the attribute to look for
        @param value: the value to look for
        '''
        num = 0
        if self.modules != None:
            for (module, state, ammo) in self.modules:
                try:
                    if module.ID != 0 and module.attributes[attribute].getModifiedValue() == value and state >= minState:
                        num = num + 1
                except KeyError:
                    pass
                
        return num
    
    def getDronesAttributeSum(self, attribute, activeOnly):
        '''
        Get the sum of an attribute on all drones, this can be used to get the combined resource usage of drones
        @param attribute: the attribute to get the sum for
        '''
        Sum = 0
        if self.drones != None:
            for drone, (amount, active) in self.drones.iteritems():
                try:
                    if activeOnly: amount = active
                    Sum = Sum + drone.attributes[attribute].getModifiedValue() * amount
                except KeyError:
                    pass
                
        return Sum    
    
    def getModuleAttributeSum(self, attribute, minState = STATE_OFFLINE):
        '''
        Get the sum of an attribute on all modules, this can be used to get the combined resource usage of modules
        @param attribute: the attribute to get the sum for
        '''
        Sum = 0
        if self.modules != None:
            for (module, state, ammo) in self.modules:
                try:
                    if state >= minState:
                        Sum = Sum + (module.getModifiedAttribute(attribute) or 0)
                except KeyError:
                    pass
                
        return Sum    
    
    def calculateCapUsage(self):
        '''
        Calculates the capacitor recharge rate and puts it in _peakCapRecharge
        '''
        capacity = self.ship.attributes["capacitorCapacity"].getModifiedValue()
        rechargeTime = self.ship.attributes['rechargeRate'].getModifiedValue() / 1000.0
        #Let's calculate cap usage now
        sum = 0
        for (module, state, ammo) in self.modules:
            try:
                if state >= STATE_ACTIVE:
                    if "speed" in module.attributes:
                        speed = module.attributes["speed"].getModifiedValue() / 1000.0
                    else:
                        speed = module.attributes["duration"].getModifiedValue() / 1000.0
                        
                    capNeed = module.attributes["capacitorNeed"].getModifiedValue()
                    capUsage = - capNeed / speed
                    sum = sum + capUsage
                    module.attributes["_capUsage"] = attribute.basicAttribute(module, "_capUsage", None, capUsage)
            except KeyError:
                pass

        #Figure stability / discharge time
        capDrain = self.ship.getModifiedAttribute("_drain") or 0
        dischargeRate = -self.getModuleAttributeSum("_capUsage", STATE_ACTIVE) - capDrain
        capBoost = self.getModuleAttributeSum("_capBoost", STATE_ACTIVE) + (self.ship.getModifiedAttribute("_capBoost") or 0)
        peakRecharge = calcCapRechForGivenCap(capacity * PEAK_RECHARGE, capacity, rechargeTime) + capBoost
        self.ship.attributes['_peakCapRecharge'] = attribute.basicAttribute(self.ship, "_peakCapRecharge", None, peakRecharge)
        currRecharge = 0
        currCap = capacity
        if peakRecharge > dischargeRate:
            low = PEAK_RECHARGE
            high = 1.0
            capStable = True
            while True:
                oldRecharge = currRecharge
                mid = (low + high) / 2.0
                currRecharge = calcCapRechForGivenCap(capacity * mid, capacity, rechargeTime) + capBoost
                diff = abs(oldRecharge - currRecharge)
                if diff <= 0.001:
                    capState = mid * 100
                    break
                else:
                    if currRecharge > dischargeRate:
                        low = mid
                    else:
                        high = mid
        else:
            capStable = False
            for i in itertools.count(1):
                oldCap = currCap
                rechargeRate = calcCapRechForGivenCap(currCap, capacity, rechargeTime) + capBoost
                currCap = currCap - dischargeRate + rechargeRate
                if currCap <= 0:
                    capState = i
                    break
        
        self.ship.attributes["_capStable"] = attribute.basicAttribute(self.ship, "_capStable", None, capStable)
        self.ship.attributes["_capState"] = attribute.basicAttribute(self.ship, "_capState", None, capState)
            
    def calculateTankEhp(self, damagePattern = defaultDamagePattern):
        '''
        Calculates the effective HP of this fit and put it in _tankEhp
        @param damagePattern: the damage pattern attackers are using
        '''
        tankEhp = 0
        #shield first, then armor, then structure
        totalDamage = float(sum(damagePattern.values()))
        for (prefix, attr) in [('shield', 'shieldCapacity'), ('armor', 'armorHP'), ('', 'hp')]:
            rawCapacity = self.ship.attributes[attr].getModifiedValue()
            specificDivider = 0
            for damageType, amount in damagePattern.iteritems():
                if prefix != '':
                    damageType = damageType[0].capitalize() + damageType[1:]
                
                resonance = self.ship.attributes[prefix + damageType + 'DamageResonance'].getModifiedValue()
                specificDivider += amount / totalDamage * resonance
            
            if specificDivider == 0: specificDivider = 1    
            currentLayerEhp = rawCapacity / specificDivider
            
            attrName = "_" + prefix.capitalize() + "Ehp"
            self.ship.attributes[attrName] = attribute.basicAttribute(self.ship, attrName, None, currentLayerEhp)
            tankEhp = tankEhp + currentLayerEhp
        
        self.ship.attributes['_totalEhp'] = attribute.basicAttribute(self.ship, '_totalEhp', None, tankEhp)
        
    def calculateTankRegen(self, damagePattern = defaultDamagePattern):
        '''
        Calculate how much DPS can be tanked by this fit and put it in _tankRegen
        @param damagePattern: the damage pattern attackers are using
        '''
        #Passive shields
        rechargeTime = self.ship.attributes['shieldRechargeRate'].getModifiedValue() / 1000.0
        capacity = self.ship.attributes['shieldCapacity'].getModifiedValue()
        #Shield follows the same rule as cap does
        rawRecharge = calcCapRechForGivenCap(capacity * PEAK_RECHARGE, capacity, rechargeTime)
        self.ship.attributes["_passiveShieldRawRecharge"] = attribute.basicAttribute(self.ship, "_passiveShieldRawRecharge", None, rawRecharge)
        pShieldRegen = self.calculateSpecificRecharge("shield", rawRecharge, damagePattern)
        self.ship.attributes["_passiveShieldRegen"] = attribute.basicAttribute(self.ship, "_passiveShieldRegen", None, pShieldRegen)
        for (tankType, resonanceType) in (("shield", "shield"), ("armor", "armor"), ("hull", "")):
            rawAttrName = "_" + tankType + "RawRecharge"
            attrName = "_" + tankType + "Regen"
            specificRecharge = 0
            for module, state, ammo in self.modules:
                if state >= STATE_ACTIVE:
                    rawRecharge = module.getModifiedAttribute(rawAttrName)
                    if rawRecharge == None: continue
                    modSpecificRecharge = self.calculateSpecificRecharge(resonanceType, rawRecharge, damagePattern)
                    module.attributes[attrName] = attribute.basicAttribute(module, attrName, None, modSpecificRecharge)
                    specificRecharge += modSpecificRecharge
            shipRawRecharge = self.ship.getModifiedAttribute(rawAttrName) or 0
            shipSpecificRecharge = self.calculateSpecificRecharge(resonanceType, shipRawRecharge, damagePattern)
            specificRecharge += shipSpecificRecharge
            self.ship.attributes[attrName] = attribute.basicAttribute(self.ship, attrName, None, specificRecharge)
    
    def calculateSpecificRecharge(self, prefix, rawRecharge, damagePattern):
        totalDamage = float(sum(damagePattern.values()))
        specificDivider = 0
        for damageType, amount in damagePattern.iteritems():
            if prefix: damageType = damageType.capitalize()
            resonance = self.ship.getModifiedAttribute(prefix + damageType + 'DamageResonance')
            specificDivider += amount / totalDamage * resonance
        
        if specificDivider == 0: specificDivider = 1
        return rawRecharge / specificDivider
                    
    def calculateSustainedTankRegen(self):
        reppers = []
        recharge = self.ship.getModifiedAttribute("_peakCapRecharge")
        capLeft = recharge + self.getModuleAttributeSum("_capUsage", STATE_ACTIVE)
        #Figure all the modules that are useful
        for tankType in ("shield", "armor", "hull"):
            reinfAttrName = "_" + tankType + "Regen"
            for module, state, ammo in self.modules:
                if state >= STATE_ACTIVE:
                    reinf = module.getModifiedAttribute(reinfAttrName)
                    if reinf:
                        cap = -module.getModifiedAttribute("_capUsage")
                        capLeft += cap
                        reppers.append((tankType, reinf, cap))
                        
        reppers.sort(cmp = lambda x, y: cmp(y[1] / y[2], x[1] / x[2]))
        #Now we have a sorted list of reppers, go through them
        sustainedRegen = {'shield' : 0,
                          'armor' : 0,
                          'hull' : 0}
        for type, reinf, cap in reppers:
            if capLeft <= 0:
                break
            
            activity = min(1, capLeft / cap)
            capLeft -= cap
            sustainedRegen[type] += activity * reinf
        for type, regen in sustainedRegen.iteritems():
            attrName = "_" + type + "SustainedRegen"
            self.ship.attributes[attrName] = \
                attribute.basicAttribute(self.ship, attrName, None, regen, 1)
        
    def calculatePerSecond(self, target, containerName): 
        '''
        Calculate the dps of this fit based on it's volley information and puts it in _dps.
        Ensure volley data is correct before calling this function
        '''
        if self.modules != None:
            for (module, state, ammo) in self.modules:
                #The try block will ensure modules that are not weapons will just get ignored
                try:
                    rof = (module.getModifiedAttribute("speed") or module.getModifiedAttribute("duration")) / 1000.0
                    delay = module.getModifiedAttribute("moduleReactivationDelay")
                    if delay and delay > rof:
                        rof = delay / 1000.0
                    dps = module.getModifiedAttribute(target) / rof
                    module.attributes[containerName] = attribute.basicAttribute(module, containerName, None, dps)
                except (KeyError, TypeError):
                    pass
        
        if self.drones != None:
            for drone, (amount, active) in self.drones.iteritems():
                try:
                    rof = drone.attributes['speed'].getModifiedValue() / 1000.0
                    dps = drone.attributes[target].getModifiedValue() / rof
                    drone.attributes[containerName] = attribute.basicAttribute(drone, containerName, None, dps)
                except: pass
                    
    def calculateVolley(self, targetResistance = { 'thermal' : 0, 'kinetic' : 0, 'explosive' : 0, 'em' : 0}):
        '''
        Calculate volley information of this fit and puts it in _volley
        @param targetResistance: the resistance the target of attacks has
        '''
        if self.modules != None:
            for (module, state, ammo) in self.modules:
                if state >= STATE_ACTIVE:
                    #The try block ensures that modules without ammo or modules that are not weapon don't crash the whole thing
                    try:
                        volley = self.getVolleyModule(module, ammo, targetResistance)
                        module.attributes['_volley'] = attribute.basicAttribute(module, '_volley', None, volley)
                    except (KeyError, AttributeError):
                        pass

        if self.drones != None:
            for drone, (amount, active) in self.drones.iteritems():
                volley = self.getVolleyModule(drone, drone, targetResistance)
                drone.attributes['_volley'] = attribute.basicAttribute(drone, '_volley', None, volley)

    def getVolleyModule(self, module, ammo, targetResistance):
        damageMultiplier = module.getModifiedAttribute("damageMultiplier") or 1
        damage = {}
        for damageType in targetResistance.iterkeys():
            try:
                damage[damageType] = ammo.getModifiedAttribute(damageType + 'Damage') or 0
            except AttributeError:
                damage[damageType] = module.getModifiedAttribute(damageType + 'Damage') or 0
                
        volley = 0
        for type, amount in damage.iteritems():
            volley = volley + amount * (1 - targetResistance[type]) * damageMultiplier
            
        return volley
    
    def calculateAlignTime(self):
        mass = self.ship.getModifiedAttribute("mass")
        agility = self.ship.getModifiedAttribute("agility")
        alignTime = -math.log(0.25) * mass *  agility / 1000000.0
        self.ship.attributes["_alignTime"] = attribute.basicAttribute(self.ship, '_alignTime', None, alignTime)
    
    def calculateLockTime(self, signature):
        scanRes = self.ship.getModifiedAttribute("scanResolution")
        if scanRes == 0: return 0
        else:
            #Not all versions of python have asinh available
            try:
                return 40000.0 / (scanRes * math.pow(math.asinh(signature), 2))
            except:
                return 0
            
    def calculateModifiedAttributes(self, calculatedFits = None):
        '''
        modify the attributes of this ship with the effects of the ship itself, the modules on it,
        the ammo in the modules, the character's skills and the implants on the character
        '''
        #Clear blocked modules
        self.blockedModules.clear()
        #Runtime dict to store when to call what
        order = []
        runTime = {}
        calculatedFits = calculatedFits or [self]
        self.gangModules.clear()
        def processFit(fit):
            if fit not in calculatedFits:
                calculatedFits.append(fit)
                fit.calculateModifiedAttributes(calculatedFits)
                
        def addRuntime(name):
            order.append(name)
            runTime[name] = {"late": [],
                              "normal": [],
                              "early" : []}
        
        def addToRuntime(name, prio, stuff):
            runTime[name][prio or "normal"].append(stuff)
            
        #Skills stuff
        addRuntime("skill")
        if self.character:
            if self.character.skills:
                for skill, level in self.character.skills.iteritems():
                    skill.clearModifiedAttributes()
                    for skillEffect in skill.effects:
                        if skillEffect.func and "normal" in skillEffect.type:
                            addToRuntime("skill", skillEffect.runTime,
                                         (skillEffect, {"level" : level}))
        
        addRuntime("ship")
        self.ship.clearModifiedAttributes()
        #Ship hull effects
        for shipEffect in self.ship.effects:
            if shipEffect.func:
                if "normal" in shipEffect.type:
                    addToRuntime("ship", shipEffect.runTime, (shipEffect, None))
                elif "gang" in shipEffect.type:
                    self.gangModules[self.ship] = {}
        
        #Loop through drones
        addRuntime("drone")
        if self.drones:
            for drone in self.drones.iterkeys():
                drone.clearModifiedAttributes()
                for droneEffect in drone.effects:
                    if droneEffect.func and "normal" in droneEffect.type:
                        addToRuntime("drone", droneEffect.runTime, (droneEffect, None))
                        
        #Loop through boosters
        addRuntime("booster")
        for booster, activeSideEffects in self.boosters.itervalues():
            booster.clearModifiedAttributes()
            for boosterEffect in booster.effects:
                if boosterEffect.func and \
                  (not "boosterSideEffect" in boosterEffect.type or \
                   boosterEffect.name in activeSideEffects):
                    addToRuntime("booster", boosterEffect.runTime, (boosterEffect, None))
                    
        #Figure when to call implants
        addRuntime("implant")
        for implant in self.implants.itervalues():
            implant.clearModifiedAttributes()
            for implantEffect in implant.effects:
                if implantEffect.func and "normal" in implantEffect.type:
                    addToRuntime("implant", implantEffect.runTime, (implantEffect, None))
                        
        #Figure what modules and ammo
        addRuntime("overload")
        addRuntime("ammo")
        addRuntime("module")
        if self.modules:
            for (module, state, ammo) in self.modules:
                if state >= STATE_OFFLINE:
                    module.ammo = ammo
                    if ammo:
                        ammo.clearModifiedAttributes()
                        if ammo.effects:
                            for ammoEffect in ammo.effects:
                                if ammoEffect.func and "normal" in ammoEffect.type:
                                    addToRuntime("ammo", ammoEffect.runTime,
                                                 (ammoEffect, {"containerModule" : module}))
                    
                    module.clearModifiedAttributes()
                    for modEffect in module.effects:
                        if modEffect.func:
                            name = modEffect.name
                            if "gang" in modEffect.type:
                                self.gangModules[module] = {"state" : state}
                            elif "normal" in modEffect.type or modEffect.type == ("active",):
                                addToRuntime("module", modEffect.runTime,
                                             (modEffect, {"state" : state}))
                            elif "overload" in modEffect.type:
                                addToRuntime("overload", modEffect.runTime,
                                             (modEffect, {"state": state}))
        
        #Gang stuff
        addRuntime("gang")
        chain = itertools.chain(self.gangModules.keys(), self.extraGangModules.keys())
        for module in chain:
            #Check for ondemand skills and add em
            if isinstance(module, item.item):
                extraSkills = []
                for stuffEffect in module.effects:
                    if stuffEffect.extraData and stuffEffect.extraData["extraSkills"]:
                        for thing in stuffEffect.extraData["extraSkills"]:
                            extraSkills.append(thing)
                for skill in module.requiredSkills:
                    skill.clearModifiedAttributes()
                    for skillEffect in skill.effects:
                        if (skillEffect.func and "ondemand" in skillEffect.type) \
                           or skill.name in extraSkills:
                            if not skill in self.extraGangModules:
                                self.extraGangModules[skill] = {"level" : 0}
                            #Autoactivate, no choice in the matter
                            if not skill in self.activeGangModules:
                                self.activeGangModules[skill] = {}
                            #Break so we don't end up processing this skill multiple times
                            break

        self.gangModules.update(self.extraGangModules)
        def process(module, overrides):
            if module in self.gangModules: overrides.update(self.gangModules[module])
            for gangEffect in module.effects:
                if gangEffect.func and "gang" in gangEffect.type:
                    addToRuntime("gang", gangEffect.runTime, (gangEffect, overrides))
                        
        for module, overrides in self.activeGangModules.iteritems():
            if isinstance(module, item.item):
                module.clearModifiedAttributes()
                process(module, overrides)
            elif isinstance(module, fitting):
                fit = module
                processFit(fit)
                for mod, overrides in fit.activeGangModules.iteritems():
                    if isinstance(mod, item.item):
                        process(mod, overrides)
        
        #Figure projected effects
        addRuntime("projected")
        def process(module, amount, active, overrides):
            if active:
                for projEffect in module.effects:
                    if projEffect.func and "projected" in projEffect.type:
                        addToRuntime("projected", projEffect.runTime,
                                     (projEffect, overrides, amount))
                        
        fitsToProject = []
        for module, (amount, active, overrides) in self.projEffects.iteritems():
            if isinstance(module, item.item):
                process(module, amount, active, overrides)
            elif isinstance(module, fitting) and active:
                fitsToProject.append((module, amount, active))

        for fit, amount, active in fitsToProject:
            processFit(fit)
            for (module, state, ammo) in fit.modules:
                for modEffect in module.effects:
                    if "projected" in modEffect.type:
                        process(module, amount, active, {"state": state})
                        break
                    
        for time in ("early", "normal", "late"):
            for layer in order:
                effects = runTime[layer][time]
                for stuff in effects:
                    try:
                        effect = stuff[0]
                        overrides = stuff[1]
                        if len(filter(lambda n: n != "active", effect.type)) > 1:
                            overrides["activeLayer"] = layer
                        if len(stuff) >= 2: rest = stuff[2:]
                        if layer == "projected":
                            for i in range(rest[0]):
                                effect.func(effect, self, **overrides)
                        else:
                            if overrides:
                                effect.func(effect, self, **overrides)
                            else:
                                effect.func(effect, self)
                    except Exception, e:
                        sys.stderr.write("error while processing " + effect.name + "\n")
                        traceback.print_exc()
            
        if self.factorInReload:
            self.factorReloadTime()
    
    def factorReloadTime(self):
        for module, state, ammo in self.modules:
            if ammo and module.getModifiedAttribute("_ammoConsumed"):
                if "duration" in module.attributes:
                    rateAttr = module.attributes["duration"]
                else:
                    rateAttr = module.attributes["speed"]
                rate = rateAttr.getModifiedValue()
                chargeSize = ammo.getModifiedAttribute("volume")
                containerSize = module.getModifiedAttribute("capacity")
                cyclesTillEmpty = int(containerSize / chargeSize)
                emptyTime = cyclesTillEmpty * rate + 10000
                rate = emptyTime / cyclesTillEmpty
                rateAttr.overrideValue = rate
    
    def block(self, module): self.blockedModules.add(module)
        
    def disableImpossibleMods(self):
        groups = {}
        changes = False
        for i in range(len(self.modules)):
            module, state, ammo = self.modules[i]
            if state >= STATE_ACTIVE:
                ##Check if the module has been blocked
                if module in self.blockedModules:
                    self.modules[i] = (module, STATE_INACTIVE, ammo)
                    changes = True
                
                ##Check if the module is over it's max limit
                maxGroupActive = module.getModifiedAttribute("maxGroupActive")
                if maxGroupActive != None:
                    if not module.group.name in groups: groups[module.group.name] = 0
                    groups[module.group.name] += 1
                    if groups[module.group.name] > maxGroupActive:
                        self.modules[i] = (module, STATE_INACTIVE, ammo)
                        changes = True
                        
        return changes
    
    def canActivateModule(self, module):
        ##Check if the module has been blocked
        if module in self.blockedModules or module.ID == 0: return False
        #Check for a maxGroupActive limit
        active = 0
        maxGroupActive = module.getModifiedAttribute("maxGroupActive")
        if maxGroupActive:
            for currModule, state, ammo in self.modules:
                if currModule.ID == 0: continue
                if currModule.group.name == module.group.name and state >= STATE_ACTIVE: active += 1
        return active < maxGroupActive or maxGroupActive == None
    
    def getModuleSlot(self, module):
        #This is tricky, we can't assume the module is part of the fit, so it's effects haven't been added yet.
        #We'll calculate effects without fit and ignore all errors
        for effect in module.effects:
            try:
                effect.func(effect, None, None)
            except Exception:
                pass
        
        return module.getModifiedAttribute("_slot")
    
    def canAddModule(self, module):
        '''
        Returns wether the passed module can be fit on this fit
        @param module: the module to check for
        '''
        
        moduleSlot = self.getModuleSlot(module)
        slots = self.getSlotsLeft(moduleSlot) > 0
        if slots == False: return False
        
        #Check if the module is restricted to a certain ship
        fitsOn = module.getModifiedAttribute("fitsToShipType")
        if fitsOn != None and fitsOn != self.ship.ID: return False
        #more checking for different attributes doing the same thing
        for i in range(4):
            fitsOn = module.getModifiedAttribute("canFitShipType%d" % i)
            if fitsOn != None and fitsOn != self.ship.ID: return False
            
        #Check if the module is restricted to certain ship group(s)
        for i in range(5):
            fitsOnGroup = module.getModifiedAttribute("canFitShipGroup" + str(i + 1))
            if fitsOnGroup and fitsOnGroup != self.ship.groupID: fits = False
            else: fits = True
        if fits == False: return fits
        
        #If the mod is a subsystem, don't let two subs using the same slot fit
        subSysSlot = module.getModifiedAttribute("subSystemSlot")
        if subSysSlot:
            for (currMod, currState, currAmmo) in self.modules:
                currSysSlot = currMod.getModifiedAttribute("subSystemSlot")
                if currSysSlot == subSysSlot: return False
                
        #Check rig sizes
        rigSize = module.getModifiedAttribute("rigSize")
        if rigSize:
            rigFits = rigSize == self.ship.getModifiedAttribute("rigSize")
            if rigFits == False: return rigFits
                
        return True
    
    def addBooster(self, module, activeSideEffects = None):
        slot = module.getModifiedAttribute("boosterness")
        if slot == None: raise ValueError("Passed module is not a booster")
        self.boosters[slot] = (module, activeSideEffects or [])
        
    def addImplant(self, module):
        slot = module.getModifiedAttribute("implantness")
        if slot == None: raise ValueError("Passed module is not an implant")
        self.implants[slot] = module
        
    def addModule(self, module, ammo = None, state = 1):
        '''
        Adds a copy of this module to the fit, this does not check if it is possible to add it.
        It is assumed that the caller checks for it accordingly.
        A copy needs to be taken to ensure that the same instance of an item is not added multiple times
        @param module: the module to add
        @param ammo: the ammo to add [optional]
        '''
        #Check if we got any dummies,  if so replace it
        for i in range(len(self.modules)):
            mod = self.modules[i][0]
            if mod.ID == 0 and mod.getModifiedAttribute("_slot") == self.getModuleSlot(module):
                self.modules[i] = (copy.deepcopy(module), state, copy.deepcopy(ammo))
                return
        
        #Seems we didn't find a dummy
        self.modules.append((copy.deepcopy(module), state, copy.deepcopy(ammo)))
    
    def removeModule(self, module, ammo = None):
        for i in range(len(self.modules)):
            currModule, currState, currAmmo = self.modules[i]
            if currModule == module and currAmmo == ammo:
                del self.modules[i]
                if module in self.activeGangModules:
                    del self.activeGangModules[module]
                return
    
    def addDrone(self, drone):
        try:
            self.drones[drone] = (self.drones[drone][0] + 1, self.drones[drone][1])
        except KeyError:
            self.drones[drone] = (1, 0)
    
    def removeDrone(self, drone, removeActive):
        currDrones = self.drones[drone][0]
        if currDrones > 1:
            if removeActive: activeDrones = self.drones[drone][1] - 1
            else: activeDrones = self.drones[drone][1]
            self.drones[drone] = (currDrones - 1, activeDrones)
            return False
        else:
            del self.drones[drone]
            return True
    
    def getNumActiveDrones(self):
        total = 0
        for drone, (num, active) in self.drones.iteritems():
            if active: total += active
        return total
    
    def addAmmo(self, module, newAmmo, oldAmmo):
        for i in range(len(self.modules)):
            mod, state, ammo = self.modules[i]
            if mod == module and (ammo == oldAmmo or oldAmmo == -1):
                self.modules[i] = (mod, state, newAmmo)
                mod.ammo = newAmmo
                return True
            
        return False
                
    
    def getSlotsLeft(self, type):
        try:
            total = self.ship.getModifiedAttribute(slotNameMap[type]) or 0
            used = self.getNumModulesAttributeValue("_slot", type)
            return total - used
        except KeyError:
            return 0
        
    def getCharSkill(self, name):
        if self.character == None:
            return None, 0
        else:
            return self.character.getSkill(name)
            
    def __getstate__(self):
        modules = []
        drones = {}
        for module, state, ammo in self.modules:
            if module.ID == 0: moduleID = module.getModifiedAttribute("_slot")
            else: moduleID = module.ID
            if ammo:
                ammoID = ammo.ID
            else:
                ammoID = None
                
            modules.append((moduleID, state, ammoID))
            
        for drone, droneData in self.drones.iteritems():
            drones[drone.ID] = droneData
        
        if self.character: charName = self.character.name
        else: charName = None 
        
        extraGang = {}
        if self.extraGangModules:
            for module, overrides in self.extraGangModules.iteritems():
                if isinstance(module, item.item): extraGang[module.ID] = overrides
                else: extraGang[module] = overrides
                
        activeGang = {}
        if self.activeGangModules:
            for module, overrides in self.activeGangModules.iteritems():
                if isinstance(module, item.item): activeGang[module.ID] = overrides
                else: activeGang[module] = overrides
                
        projEffects = {}
        if self.projEffects:
            for module, args in self.projEffects.iteritems():
                if isinstance(module, item.item): projEffects[module.ID] = args
                else: projEffects[module] = args
        
        boosters = {}
        if self.boosters:
            for slot, (module, activeSideEffects) in self.boosters.iteritems():
                boosters[slot] = (module.ID, activeSideEffects)
                
        implants = {}
        if self.implants:
            for slot, module in self.implants.iteritems():
                implants[slot] = module.ID
                
        factorInReload = self.factorInReload
        
        return {'ship' : self.ship.ID,
                'modules' : modules,
                'drones' : drones,
                'extraGang' : extraGang,
                'activeGang' : activeGang,
                'projEffects' : projEffects,
                'implants' : implants,
                'boosters' : boosters,
                'factorInReload' : factorInReload}
        
    def __setstate__(self, state):
        ship = item.getItem(state["ship"])
        modules = []
        #See why we're keeping this cache a little lower, in the gang section
        mods = {}
        for (modID, stateID, ammoID) in state["modules"]:
            if isinstance(modID, str): mod = generateDummy(modID)
            else: mod = item.getItem(modID)
            mods[modID] = mod
            if ammoID: ammo = item.getItem(ammoID)
            else: ammo = None
            modules.append((mod, stateID, ammo))
            
        drones = {}
        for droneID, (amount, active) in state["drones"].iteritems():
            #Convert old format to new one
            if isinstance(active, bool):
                if active: active = amount
                else: active = 0
                
            drone = item.getItem(droneID)
            drones[drone] = (amount, active)
        
        if "extraGang" in state and "activeGang" in state:
            #We need to slap the same instance of the item in both dictionaries
            #If we don't then effects won't get all get called on the same instance
            #(that's bad, because skills affecting mods won't be able to boost correctly)
            extraGang = {}
            activeGang = {}
            for modID, overrides in state["extraGang"].iteritems():
                if isinstance(modID, int):
                    mod = item.getItem(ID = modID)
                else: mod = modID
                mods[modID] = mod
                extraGang[mod] = overrides
        
            for modID, overrides in state["activeGang"].iteritems():
                if modID in mods: mod = mods[modID]
                else: mod = item.getItem(ID = modID)
                activeGang[mod] = overrides
        else:
            extraGang = None
            activeGang = None
            
        if "projEffects" in state:
            projEffects = {}
            for modID, args in state["projEffects"].iteritems():
                if isinstance(modID, int):
                    mod = item.getItem(ID = modID)
                else: mod = modID
                projEffects[mod] = args
        else:
            projEffects = None
            
        if "factorInReload" in state:
            factorInReload = state["factorInReload"]
        else:
            factorInReload = False
    
        if "implants" in state:
            implants = {}
            for slot, modID in state["implants"].iteritems():
                mod = item.getItem(ID = modID)
                implants[slot] = mod
        else:
            implants = None
                
        if "boosters" in state:
            boosters = {}
            for slot, (modID, activeSideEffects) in state["boosters"].iteritems():
                mod = item.getItem(ID = modID)
                boosters[slot] = (mod, activeSideEffects)
        else:
            boosters = None
        self.__init__(ship, modules, None, drones, extraGang, activeGang, projEffects, implants, boosters, factorInReload)
    
    def importEft(self, eftString):
        eftString = eftString.strip()
        typeNameRe = re.compile("\\[(.*), (.*)\\]")
        lines = eftString.split("\n")
        shipType, fitName = re.match(typeNameRe, lines[0]).groups()
        ship = item.getItem(name = shipType)
        if ship == None:
            return None
        self.setShip(ship)
        
        for i in range(1, len(lines)):
            line = lines[i]
            modAmmo = line.split(",")
            modDrone = modAmmo[0].split(" x")
            if len(modAmmo) == 2: ammoName = modAmmo[1].strip()
            else: ammoName = None
            modName = modDrone[0].strip()
            if len(modDrone) == 2: droneAmount = modDrone[1].strip()
            else: droneAmount = None
            mod = item.getItem(name=modName)
            ##Temporary workaround for EFT not having new rig names
            if not mod: mod = item.getItem(name="Large " + modName)
            ##Workaround for items with " x" in them
            if not mod: mod = item.getItem(name= modAmmo[0])
            if mod:
                if ammoName:
                    ammo = item.getItem(name=ammoName)
                else:
                    ammo = None
                if mod.group.category.name == "Drone":
                    if droneAmount is None: droneAmount = 1
                    droneAmount = int(droneAmount)
                    self.drones[mod] = (droneAmount, False)
                else: self.modules.append((mod, STATE_ACTIVE, ammo))
                
        return fitName
    
    def exportEft(self, name):
        export = "[%s, %s]\n" % (self.ship.name, name)
        stuff = {}
        for module, state, ammo in self.modules:
            slot = self.getModuleSlot(module)
            if not slot in stuff: stuff[slot] = []
            curr = module.name
            if ammo:
                curr += ", %s" % ammo.name
            curr += "\n"
            stuff[slot].append(curr)
        
        for slot in slotTypes:
            if slot in stuff:
                export += "\n"
                for curr in stuff[slot]:
                    export += curr
                    
        export += "\n\n"
        for drone, (amount, active) in self.drones.iteritems():
            export += "%s x%s\n" % (drone.name, amount)
            
        return export

    def exportXml(self, name):        
        doc = xml.dom.minidom.Document()
        fittings = doc.createElement("fittings")
        doc.appendChild(fittings)
        fitting = doc.createElement("fitting")
        fitting.setAttribute("name", name)
        fittings.appendChild(fitting)
        description = doc.createElement("description")
        description.setAttribute("value", "")
        fitting.appendChild(description)
        shipType = doc.createElement("shipType")
        shipType.setAttribute("value", self.ship.name)
        fitting.appendChild(shipType)
        
        slotNum = {}
        for module, state, ammo in self.modules:
            slot = slotShortNameMap[self.getModuleSlot(module)]
            if not slot in slotNum: slotNum[slot] = 0
            slotId = slotNum[slot]
            slotNum[slot] += 1
            hardware = doc.createElement("hardware")
            hardware.setAttribute("type", module.name)
            hardware.setAttribute("slot", "%s slot %d" % (slot, slotId))
            fitting.appendChild(hardware)
            
        for drone, (amount, active) in self.drones.iteritems():
            hardware = doc.createElement("hardware")
            hardware.setAttribute("qty", "%d" % amount)
            hardware.setAttribute("slot", "drone bay")
            hardware.setAttribute("type", drone.name)
            fitting.appendChild(hardware)
            
        return doc.toprettyxml()
    
    def importXml(self, text):
        doc = xml.dom.minidom.parseString(text)
        fittings = doc.getElementsByTagName("fittings").item(0)
        fitting = fittings.getElementsByTagName("fitting").item(0)
        name = fitting.getAttribute("name")
        shipType = fitting.getElementsByTagName("shipType").item(0)
        shipType = shipType.getAttribute("value")
        ship = item.getItem(name = shipType)
        self.setShip(ship)
        hardwares = fitting.getElementsByTagName("hardware")
        for hardware in hardwares:
            moduleName = hardware.getAttribute("type")
            quantity = hardware.getAttribute("qty")
            mod = item.getItem(name=moduleName)
            if mod.group.category.name == "Drone":
                self.drones[mod] = (int(quantity), 0)
            else:
                self.modules.append((mod, STATE_ACTIVE, None))
                
        return name
    
def generateDummy(slot):
    name = slot.capitalize() + " Slot"
    dummy = item.item(0, name, name)
    dummy.baseAttributes = {}
    dummy.attributes["_slot"] = attribute.basicAttribute(dummy, "_slot", slot, None)
    return dummy
    
def ammoFits(module, ammo):
    '''
    Wether the passed ammo fits in the passed module
    @param module: the module
    @param ammo: the ammo
    '''
    try:
        #Check sizes, if ammo size > module volume it won't fit
        chargeVolume = ammo.getModifiedAttribute("volume")
        moduleCapacity = module.getModifiedAttribute("capacity")
        if chargeVolume > moduleCapacity: return False
        
        chargeSize = module.getModifiedAttribute("chargeSize")
        if chargeSize != None:
            ammoSize = ammo.getModifiedAttribute('chargeSize')
            if chargeSize != ammoSize:
                return False

        ammoGroup = ammo.groupID
        for i in range(5):
            chargeGroup = module.getModifiedAttribute('chargeGroup' + str(i))
            if chargeGroup == None:
                continue
            
            if ammoGroup == chargeGroup:
                return True
            
    except KeyError:
        pass
    
    return False

#Make sure to return the same dict across calls
savedData = None
def loadFits():
    '''
    Load fitting data from hard disk (location: ~/.pyfa/fits)
    '''
    global savedData
    if savedData == None:
        packedData = dataFolder.pickleFileToObj("fits")
        savedData = {}
        if packedData:
            for groupName, ships in packedData.iteritems():
                savedData[groupName] = {}
                for shipID, fits in ships.iteritems():
                    ship = item.getItem(ID = shipID)
                    savedData[groupName][ship] = {}
                    for fitName, fit in fits.iteritems():
                        savedData[groupName][ship][fitName] = fit
        
    return savedData

def saveFits(fits):
    '''
    Save fitting data to hard disk (location: ~/.pyfa/fits)
    '''
    packedFits = {}
    for groupName, ships in fits.iteritems():
        packedFits[groupName] = {}
        for ship, fits in ships.iteritems():
            packedFits[groupName][ship.ID] = {}
            for fitName, fit in fits.iteritems():
                packedFits[groupName][ship.ID][fitName] = fit
                
    return dataFolder.pickleObjToFile(packedFits, "fits")

def calcCapRechForGivenCap(capacitorState, capacitorCapacity, rechargeTime):
    '''
    Calculate the capacitor recharge rate for the passed capacitor state.
    @param capacitorState: the capacitor's current charge
    @param capacitorCapacity: the capacitor's capacity
    @param rechargeTime: the capacitor's recharge time
    '''
    return capacitorCapacity * ((4.9678 / rechargeTime) *
              (1 - (capacitorState / capacitorCapacity)) *
              math.sqrt((2 * (capacitorState / capacitorCapacity)) -
                        pow((capacitorState / capacitorCapacity), 2)))
